/*
 * This file is part of SkunkDB
 * Copyright (c) 2002 Gianni Tedesco
 * Released under the terms of the GNU GPL version 2
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>

#include <mpool.h>
#include <fbuf.h>
#include <avl.h>

/* Map in the database */
struct avl_db *avl_open(int fd)
{
	struct avl_hdr *h;
	struct avl_db *d;
	struct stat st;

	if ( fstat(fd, &st) )
		return NULL;

	if ( st.st_size < sizeof(struct avl_hdr) )
		return NULL;

	if ( !(d=calloc(1, sizeof(*d))) )
		return NULL;

	if ( (d->map=mmap(NULL, st.st_size, PROT_READ,
		MAP_SHARED, fd, 0))==MAP_FAILED ) {
		free(d);
		return NULL;
	}

	/* Check the header */
	h=(struct avl_hdr *)d->map;
	if ( h->magic != AVL_MAGIC ) {
		munmap(d->map, st.st_size);
		free(d);
		return NULL;
	}

	d->count=h->count;
	d->maplen=st.st_size;
	d->end=d->map + st.st_size;
	d->r=(struct avl_disk *)(d->map + sizeof(struct avl_hdr));

	/* Check the tree nodes are in */
	if ( (char *)&d->r[d->count] > d->end ) {
		munmap(d->map, st.st_size);
		free(d);
		return NULL;
	}

	/* Find the range of keys in this DB */
	d->first=d->r[0].key;
	d->last=d->r[d->count-1].key;

	/* Closing the file does not unmap it, we
	 * just close it to save resources, heh */
	close(fd);

	return d;
}

#if 1
#define dprintf(x...)
#else
#define dprintf printf
#endif

int avl_select(struct avl_db *d, unsigned int first, unsigned int last)
{
	size_t start=0, end=d->count;
	size_t estart=0, eend=d->count;
	avl_idx_t *f, *l, *t;
	size_t fi=0, li=0;
	size_t i;

	/* Clamp search parameters to actual range of keys */
	if ( first < d->first )
		first=d->first;

	if ( last > d->last )
		last=d->last;

	/* Sanity check */
	if ( last < first )
		return -1;

	/* Binary search for the lowest index */
	for (;;) {
		i=start;
		i+=(end-start)/2;

		dprintf("%u-%u : %u is %u\n",
			start, end, i, d->r[i].key);
	
		if ( start+1 >= end ) {
			fi=i;
			break;
		}

		/* Keep track of roughly where the last index is */
		if ( d->r[i].key > last ) {
			eend=i;
		}
		if ( estart==0 && d->r[i].key < last ) {
			estart=i;
		}

		if ( d->r[i].key>first ) {
			end = i - 1;
			continue;
		}

		if ( d->r[i].key<first ) {
			start = i + 1;
			continue;
		}

		fi=i;
		break;
	}

	dprintf("search from indexes %u to %u\n", estart, eend);

	/* Tighten down on finding that last index */
	for (;;) {
		i=estart;
		i+=(eend-estart)/2;

		dprintf("%u-%u : %u is %u\n",
			estart, eend, i, d->r[i].key);
	
		if ( estart+1 >= eend ) {
			li=i;
			break;
		}

		if ( d->r[i].key>last ) {
			eend = i - 1;
			continue;
		}

		if ( d->r[i].key<last ) {
			estart = i + 1;
			continue;
		}

		li=i;
		break;
	}

	dprintf("selected records %u to %u\n", fi, li);

	/* Seeing as records are stored in order we can
	 * get the address of the first, then the last */
	f=(avl_idx_t *)(d->map + d->r[fi].ofs);
	l=(avl_idx_t *)(d->map + d->r[li].ofs)
		+ (d->r[li].len - 1);

	/* check the results are inside our mapping */
	if ( (char *)l > d->end )
		return -1;

	/* Display the results */
	printf("mapped %u keys to %u values\n", (li+1)-fi, (l+1)-f);
	for(t=f; t<=l; t++) {
		printf("%u\n", *t);
	}

	return 0;
}

/* Close the database */
void avl_close(struct avl_db *d)
{
	munmap(d->map, d->maplen);
	free(d);
}
