/*
 * This file is part of SkunkDB
 * Copyright (c) 2002 Gianni Tedesco
 * Released under the terms of the GNU GPL version 2
 *
 * avl tree maintenance functions taken from avl-0.4, original copyright
 * notice is reproduced below:
 *
 * (C) 2000 Daniel Nagy, Budapest University of Technology and Economics
 * Released under GNU General Public License (GPL) version 2
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <mpool.h>
#include <fbuf.h>
#include <avl.h>

#define AVL_BATCH 32768

struct avl_make *cur_db;

struct avl_make *avl_new_db(int fd)
{
	struct avl_make *d;

	if ( fd<0 )
		return NULL;

	if ( !(d=calloc(1, sizeof(*d))) ) 
		return NULL;

	d->fd=fd;

	mpool_new(&d->rec,
		sizeof(struct avl),
		AVL_BATCH);

	return d;
}

/* swing left */
static inline void avl_swl(struct avl **r)
{
	struct avl *a=*r;
	struct avl *b=a->right;
	*r=b;
	a->right=b->left;
	b->left=a;
}

/* swing right */
static inline void avl_swr(struct avl **r)
{
	struct avl *a=*r;
	struct avl *b=a->left;
	*r=b;
	a->left=b->right;
	b->right=a;
}

/* balance maintainance */
static inline void avl_balance(struct avl *r)
{
	switch ( r->balance ) {
	case -1:
		r->left->balance=0;
		r->right->balance=1;
		break;
	case 1:
		r->left->balance=-1;
		r->right->balance=0;
		break;
	case 0:
		r->left->balance=0;
		r->right->balance=0;
	}
	r->balance=0;
}

/* Recursive AVL tree insertion */
int avl_insert_key(struct avl **r, struct avl *n)
{
	struct avl *t=*r;

	if ( n->key < t->key ) {
		if ( t->left ) {
			if ( avl_insert_key(&t->left, n) ) {
				switch ( t->left->balance-- ) {
				case 1: return 0;
				case 0: return 1;
				}
				if ( t->left->balance<0 ) {
					/* simple rotate */
					avl_swr(r);
					(*r)->balance=0;
					(*r)->right->balance=0;
				}else{
					/* double rotate */
					avl_swl(&t->left);
					avl_swr(r);
					avl_balance(*r);
				}
			}
			return 0;
		}else{
			t->left=n;
			if ( t->balance-- ) return 0;
			return 1;
		}
	}

	if ( n->key > t->key ) {
		if ( t->right ) {
			if ( avl_insert_key(&t->right, n) ) {
				switch ( t->right->balance-- ) {
				case 1: return 0;
				case 0: return 1;
				}
				if ( t->right->balance<0 ) {
					/* simple rotate */
					avl_swl(r);
					(*r)->balance=0;
					(*r)->left->balance=0;
				}else{
					/* double rotate */
					avl_swr(&t->right);
					avl_swl(r);
					avl_balance(*r);
				}
			}
			return 0;
		}else{
			t->right=n;
			if ( t->balance-- ) return 0;
			return 1;
		}
	}

	/* Add duplicates to a list */
	if ( n->key==t->key ) {
		t->count++;
		n->next=t->next;
		t->next=n;
		cur_db->count--;
	}

	return 0;
}

int avl_insert(struct avl_make *d,
	u_int32_t key, avl_idx_t idx)
{
	struct avl *n;

	if ( !(n=mpool_alloc(&d->rec)) )
		return -1;

	n->left=NULL;
	n->right=NULL;
	n->key=key;
	n->balance=0;

	n->idx=idx;
	n->count=1;
	n->next=NULL;

	/* Special case for root */
	if ( !d->root ) {
		d->root=n;
		d->count++;
		return 0;
	}

	cur_db=d;
	cur_db->count++;
	avl_insert_key(&d->root, n);
	return 0;
}

/* Traverse the tree in-order and dump the nodes */
int avl_dump_nodes(struct avl_make *d, struct avl *n)
{
	if ( n->left && 
		avl_dump_nodes(d, n->left)<0 )
		return -1;

	/* Write the node out to disk */
	do {
		struct avl_disk o={
			.key=n->key,
			.ofs=d->cur,
			.len=n->count,
		};

		if ( fbuf_write(&d->buf, &o, sizeof(o))<0 )
			return -1;
	}while(0);

	d->cur += n->count * sizeof(avl_idx_t);

	if ( n->right &&
		avl_dump_nodes(d, n->right)<0 )
		return -1;

	return 0;
}

/* Traverse the tree in-order and dump the data */
int avl_dump_data(struct avl_make *d, struct avl *n)
{
	if ( n->left &&
		avl_dump_data(d, n->left)<0 )
		return -1;

	do {
		struct avl *a;
		for( a=n; a; a=a->next )
			if ( fbuf_write(&d->buf, &a->idx, sizeof(a->idx))<0 )
				return -1;
	}while(0);

	if ( n->right &&
		avl_dump_data(d, n->right)<0 )
		return -1;

	return 0;
}

int avl_commit(struct avl_make *d)
{
	struct avl_hdr h;
	int ret=0;

	if ( !d->root )
		goto finish;

	/* Work out where the first byte of data should go */
	d->cur=sizeof(struct avl_hdr) +
		(d->count * sizeof(struct avl_disk));

	/* Setup the output buffer */
	fbuf_attach(&d->buf, d->fd, 23768);

	/* Write the file header */
	h.magic=AVL_MAGIC;
	h.count=d->count;
	if ( fbuf_write(&d->buf, &h, sizeof(h))<0 ) {
		ret=-1;
		goto finish;
	}

	/* Dump the tree */
	if ( avl_dump_nodes(d, d->root)<0 ) {
		ret=-1;
		goto finish;
	}

	/* Dump the data */
	if ( avl_dump_data(d, d->root)<0 ) {
		ret=-1;
		goto finish;
	}

finish:
	if ( fbuf_close(&d->buf) )
		ret=-1;

	mpool_free(&d->rec);
	free(d);
	return ret;
}

int avl_abort(struct avl_make *d)
{
	mpool_free(&d->rec);
	free(d);
	return 0;
}
